/*
 * Core configuration class
 */

#include <fstream>
#include <iostream>
#include <string>

#include <arc/configcore.h>
#include <arc/notify.h>
#include <arc/ngconfig.h>
#include <arc/xmlconfig.h>

#ifdef HAVE_LIBINTL_H
#include <libintl.h>
#define _(A) dgettext("arclib", (A))
#else
#define _(A) (A)
#endif


void Config::AddConfValue(std::string path, std::string value) {
	Config::content[path].push_back(value);
}


OptionValue Config::ConfValue(std::string key) {

	Options::iterator i = content.find(key);
	if (i != content.end())
		return i->second;
	else
		return OptionValue();
}


std::string Config::FirstConfValue(std::string key) {
	OptionValue optval = ConfValue(key);
	if (!optval.empty())
		return optval.front();
	else
		return std::string();
}


Options& Config::GetOptions() {
	return content;
}


int Config::Size() {
	return content.size();
}


Config ReadConfig(std::istream& is) {

	try {
		notify(VERBOSE) << _("Parsing stream as xml format") << std::endl;
		is.seekg(0, std::ios::beg);
		return XMLConfig().Read(is);
	}
	catch(ConfigError) {}
	try {
		notify(VERBOSE) << _("Parsing stream as old format") << std::endl;
		is.seekg(0, std::ios::beg);
		return NGConfig().Read(is);
	}
	catch(ConfigError) {}
	throw ConfigError(_("Unknown configuration format"));
}


Config ReadConfig(const std::string& filename) {

	notify(DEBUG) << _("Reading configuration file") << ": " << filename
	              << std::endl;

	std::ifstream is(filename.c_str());
	Config conf = ReadConfig(is);
	is.close();
	return conf;
}
