// TODO: Run multiple clients in parallel

#include "../../std.h"
#include <globus_common.h>
#include <globus_io.h>
#include <string>
#include <list>
#include <fstream>

#include "../../files/info_types.h"
#include "../../misc/log_time.h"
#include <arc/stringconv.h>
#include <arc/url.h>
#include "../../misc/globus_error_utils.h"

#include "../client/client.h"
#include "../../misc/url_options.h"
#include "../../config/conf.h"

#include "logger_soap.nsmap"
#include "logger2_soap.nsmap"


std::string* soap_new_std__string(struct soap* sp,const char* s) {
  if(!s) return NULL;
  std::string* str = sp?soap_new_std__string(sp,-1):(new std::string);
  if(!str) return NULL;
  *str=s;
  return str;
}

std::string* soap_new_std__string(struct soap* sp,const std::string& s) {
  return soap_new_std__string(sp,s.c_str());
}

time_t* soap_new_time(struct soap* sp,const char* s) {
  mds_time t;
  t=s;
  if(!t.defined()) return NULL;
  time_t* p = sp?(time_t*)soap_malloc(sp,sizeof(time_t)):(new time_t);
  if(p) *p=(time_t)t;
  return p;
}

time_t* soap_new_time(struct soap* sp,const std::string& s) {
  return soap_new_time(sp,s.c_str());
}

template<typename T> T* soap_new_i(struct soap* sp,const std::string& s) {
  try {
    T n = stringto<T>(s);
    T* p = sp?(T*)soap_malloc(sp,sizeof(T)):(new T);
    if(p) *p=n;
    return p;
  } catch(std::exception e) {
    return NULL;
  };
}

template<typename T> T* soap_new_i(struct soap* sp,const char* s) {
  try {
    T n = stringto<T>(std::string(s));
    T* p = sp?soap_malloc(sp,sizeof(T)):malloc(sizeof(T));
    if(p) *p=n;
    return p;
  } catch(std::exception e) {
    return NULL;
  };
}

void copy_jobinfo(nl2__UsageRecord& to,const nl__jobinfo& from,struct soap* sp) {
  to.projectname=NULL;
  to.requestedcputime=NULL;
  to.requestedmemory=NULL;
  to.requesteddisk=NULL;
  to.localuser=NULL;
  to.queue=NULL;
  to.lrms=NULL;
  to.localjobid=NULL;
  to.lrmssubmissiontime=NULL;
  to.lrmsendtime=NULL;
  to.nodename=NULL;
  to.nodecount=NULL;
  to.usedwalltime=NULL;
  to.useddisk=NULL;
  to.status=NULL;
  to.downloadtime=NULL;
  to.uploadtime=NULL;
  if(from.start) to.submissiontime=(time_t*)soap_malloc(sp,sizeof(time_t));
  if(to.submissiontime) *(to.submissiontime)=from.start;
  if(from.end) to.endtime=(time_t*)soap_malloc(sp,sizeof(time_t));
  if(to.endtime) *(to.endtime)=from.end;
  to.cluster=from.cluster?soap_new_std__string(sp,-1):NULL;
  if(to.cluster) *(to.cluster)=from.cluster;
  to.usersn=from.user?from.user:"";
  to.ngjobid=from.id?from.id:"";
  to.jobname=from.name?soap_new_std__string(sp,-1):NULL;
  if(to.jobname) *(to.jobname)=from.name;
  to.failurestring=from.failure?soap_new_std__string(sp,-1):NULL;
  if(to.failurestring) *(to.failurestring)=from.failure;
  to.lrms=from.lrms?soap_new_std__string(sp,-1):NULL;
  if(to.lrms) *(to.lrms)=from.lrms;
  to.queue=from.queue?soap_new_std__string(sp,-1):NULL;
  if(to.queue) *(to.queue)=from.queue;
  to.description=from.rsl?soap_new_std__string(sp,-1):NULL;
  if(to.description) *(to.description)=from.rsl;
  to.clienthost=from.ui?soap_new_std__string(sp,-1):NULL;
  if(to.clienthost) *(to.clienthost)=from.ui;
  if(from.usedcpu) to.usedcputime=(int*)malloc(sizeof(int));
  if(to.usedcputime) *(to.usedcputime)=from.usedcpu;
  if(from.usedmem) to.usedmemory=(int*)malloc(sizeof(int));
  if(to.usedmemory) *(to.usedmemory)=from.usedmem;
  if(from.end) {
    to.exitcode=(int*)malloc(sizeof(int));
    if(to.exitcode) *(to.exitcode)=from.failure?(-1):0;
  };
};

class JobRecord {
 private:
  bool valid;
  bool allocated;
 public:
  nl2__UsageRecord info;
  std::string url;
  operator bool(void) { return valid; };
  bool operator!(void) { return !valid; };
  JobRecord(std::istream& i,struct soap* sp = NULL);
  JobRecord(const JobRecord& j);
  ~JobRecord(void);
};

JobRecord::JobRecord(const JobRecord& j) {
  valid=j.valid;
  info=j.info;
  allocated=false;
}

JobRecord::JobRecord(std::istream& i,struct soap* sp):valid(false),url("") {
  allocated=(sp != NULL);
  if(sp) {
    info.soap_default(sp);
  } else {
      info.ngjobid="";
      info.usersn="";
      info.cluster=NULL;
      info.description=NULL;
      info.projectname=NULL;
      info.jobname=NULL;
      info.clienthost=NULL;
      info.localuser=NULL;
      info.queue=NULL;
      info.lrms=NULL;
      info.localjobid=NULL;
      info.nodename=NULL;
      info.failurestring=NULL;
      info.status=NULL;
      info.requestedcputime=NULL;
      info.requestedmemory=NULL;
      info.requesteddisk=NULL;
      info.nodecount=NULL;
      info.exitcode=NULL;
      info.usedcputime=NULL;
      info.usedmemory=NULL;
      info.usedwalltime=NULL;
      info.useddisk=NULL;
      info.submissiontime=NULL;
      info.lrmssubmissiontime=NULL;
      info.lrmsendtime=NULL;
      info.endtime=NULL;
      info.downloadtime=NULL;
      info.uploadtime=NULL;
  };
  for(;;) {
    if(i.fail()) goto error;
    if(i.eof()) break;
    std::string value;
    std::string key = config_read_line(i,value,'=');
    if(key=="loggerurl") { url=value; }
    else if(key=="ngjobid") { info.ngjobid=value; }
    else if(key=="usersn") { info.usersn=value; }
    else if(key=="cluster") { info.cluster=soap_new_std__string(sp,value); }
    else if(key=="description") { info.description=soap_new_std__string(sp,value); }
    else if(key=="projectname") { info.projectname=soap_new_std__string(sp,value); }
    else if(key=="jobname") { info.jobname=soap_new_std__string(sp,value); }
    else if(key=="clienthost") { info.clienthost=soap_new_std__string(sp,value); }
    else if(key=="localuser") { info.localuser=soap_new_std__string(sp,value); }
    else if(key=="queue") { info.queue=soap_new_std__string(sp,value); }
    else if(key=="lrms") { info.lrms=soap_new_std__string(sp,value); }
    else if(key=="localjobid") { info.localjobid=soap_new_std__string(sp,value); }
    else if(key=="nodename") { info.nodename=soap_new_std__string(sp,value); }
    else if(key=="failurestring") { info.failurestring=soap_new_std__string(sp,value); }
    else if(key=="status") { info.status=soap_new_std__string(sp,value); }
    else if(key=="requestedcputime") { info.requestedcputime=soap_new_i<int>(sp,value); }
    else if(key=="requestedmemory") { info.requestedmemory=soap_new_i<int>(sp,value); }
    else if(key=="requesteddisk") { info.requesteddisk=soap_new_i<int>(sp,value); }
    else if(key=="nodecount") { info.nodecount=soap_new_i<int>(sp,value); }
    else if(key=="exitcode") { info.exitcode=soap_new_i<int>(sp,value); }
    else if(key=="usedcputime") { info.usedcputime=soap_new_i<int>(sp,value); }
    else if(key=="usedmemory") { info.usedmemory=soap_new_i<int>(sp,value); }
    else if(key=="usedwalltime") { info.usedwalltime=soap_new_i<int>(sp,value); }
    else if(key=="useddisk") { info.useddisk=soap_new_i<int>(sp,value); }
    else if(key=="submissiontime") { info.submissiontime=soap_new_time(sp,value); }
    else if(key=="lrmssubmissiontime") { info.lrmssubmissiontime=soap_new_time(sp,value); }
    else if(key=="lrmsendtime") { info.lrmsendtime=soap_new_time(sp,value); }
    else if(key=="endtime") { info.endtime=soap_new_time(sp,value); }
    else if(key=="downloadtime") { info.downloadtime=soap_new_time(sp,value); }
    else if(key=="uploadtime") { info.uploadtime=soap_new_time(sp,value); }
  };
  valid=true;
error:
  return;
}

JobRecord::~JobRecord(void) {
  if(allocated) {
    if(info.cluster) delete info.cluster;
    if(info.description) delete info.description;
    if(info.projectname) delete info.projectname;
    if(info.jobname) delete info.jobname;
    if(info.clienthost) delete info.clienthost;
    if(info.localuser) delete info.localuser;
    if(info.queue) delete info.queue;
    if(info.lrms) delete info.lrms;
    if(info.localjobid) delete info.localjobid;
    if(info.nodename) delete info.nodename;
    if(info.failurestring) delete info.failurestring;
    if(info.status) delete info.status;
    if(info.requestedcputime) delete info.requestedcputime;
    if(info.requestedmemory) delete info.requestedmemory;
    if(info.requesteddisk) delete info.requesteddisk;
    if(info.nodecount) delete info.nodecount;
    if(info.exitcode) delete info.exitcode;
    if(info.usedcputime) delete info.usedcputime;
    if(info.usedmemory) delete info.usedmemory;
    if(info.usedwalltime) delete info.usedwalltime;
    if(info.useddisk) delete info.useddisk;
    if(info.submissiontime) delete info.submissiontime;
    if(info.lrmssubmissiontime) delete info.lrmssubmissiontime;
    if(info.lrmsendtime) delete info.lrmsendtime;
    if(info.endtime) delete info.endtime;
    if(info.downloadtime) delete info.downloadtime;
    if(info.uploadtime) delete info.uploadtime;
  };
}

#define print_element(name) \
  if(j.name.length()) o<<#name<<"="<<j.name<<std::endl;

#define print_optional_element(name) \
  if(j.name) o<<#name<<"="<<*(j.name)<<std::endl;

std::ostream& operator<<(std::ostream& o,const nl2__UsageRecord& j) {
  print_element(ngjobid);
  print_element(usersn);
  print_optional_element(cluster);
  print_optional_element(description);
  print_optional_element(projectname);
  print_optional_element(jobname);
  print_optional_element(clienthost);
  print_optional_element(requestedcputime);
  print_optional_element(requestedmemory);
  print_optional_element(requesteddisk);
  print_optional_element(submissiontime);
  print_optional_element(localuser);
  print_optional_element(queue);
  print_optional_element(lrms);
  print_optional_element(lrmssubmissiontime);
  print_optional_element(lrmsendtime);
  print_optional_element(nodename);
  print_optional_element(nodecount);
  print_optional_element(exitcode);
  print_optional_element(failurestring);
  print_optional_element(usedcputime);
  print_optional_element(usedmemory);
  print_optional_element(usedwalltime);
  print_optional_element(useddisk);
  print_optional_element(status);
  print_optional_element(endtime);
  print_optional_element(downloadtime);
  print_optional_element(uploadtime);
  return o;
}

class LoggerClient {
 private:
  URL* url;
  HTTP_ClientSOAP* client;
  struct soap soap;
  bool SameContact(const char* url);
  bool NewURL(const char* url);
 public:
  LoggerClient(void);
  ~LoggerClient(void);
  bool Initialized(void);
  bool Report(const char* url,std::list<nl2__UsageRecord>& info);
  bool ReportV1(const char* url,std::list<nl2__UsageRecord>& info);
  bool ReportV2(const char* url,std::list<nl2__UsageRecord>& info);
  bool Query(const char* url,const char* q,unsigned long long offset,std::list<nl2__UsageRecord>& info);
  bool QueryV1(const char* url,const char* q,unsigned long long offset,std::list<nl2__UsageRecord>& info);
  bool QueryV2(const char* url,const char* q,unsigned long long offset,unsigned long long size,std::list<nl2__UsageRecord>& info);
};


LoggerClient::LoggerClient(void):url(NULL),client(NULL) {
}

LoggerClient::~LoggerClient(void) {
  if(client) delete client;
}

bool LoggerClient::Initialized(void) {
  return (client != NULL);
}

bool LoggerClient::SameContact(const char* url_) {
  if(!url) return false;
  try {
    URL u(url_);
    return((u.Protocol() == url->Protocol()) && 
           (u.Host() == url->Host()) && 
           (u.Port() == url->Port()));
  } catch (std::exception e) {
    return false;
  };
}

bool LoggerClient::NewURL(const char* url_) {
  if(url_ == NULL) return false;
  try {
    if(url) {
      URL u(url_);
      if((u.Protocol() == url->Protocol()) &&
         (u.Host() == url->Host()) &&
         (u.Port() == url->Port())) {
        *url=u;
      } else { // reinitialize for different server
        *url=u; if(client) delete client; client=NULL;
      };
    } else { // initialize for first connection
      url=new URL(url_); if(client) delete client; client=NULL;
    };
  } catch (std::exception e) {
    delete url; url=NULL; if(client) delete client; client=NULL;
    return false;
  };
  if(client == NULL) {
    std::string soap_url = url->Protocol()+"://"+url->Host()+":"+tostring(url->Port());
    client = new HTTP_ClientSOAP(soap_url.c_str(),&soap);
    client->AddSOAPNamespaces(logger_soap_namespaces);
    client->AddSOAPNamespaces(logger2_soap_namespaces);
  } else {
    client->reset();
  };
  return true;
}

bool LoggerClient::Report(const char* url_,std::list<nl2__UsageRecord>& info) {
  // Try V2, if it fails try V1
  if(ReportV2(url_,info)) return true;
  return ReportV1(url_,info);
}

bool LoggerClient::ReportV2(const char* url_,std::list<nl2__UsageRecord>& info) {
  if(!NewURL(url_)) return false;
  if(info.size() <= 0) return true;
  if(client->connect() != 0) {
    odlog(1)<<"Failed to connect to "<<*url<<std::endl;
    return false;
  };
  nl2__addRequest req; req.soap_default(&soap);
  for(std::list<nl2__UsageRecord>::iterator i = info.begin();
                                             i!=info.end();++i) {
    req.job.push_back(&(*i));
  };
  int soap_err = SOAP_OK;
  nl2__addResponse r; r.soap_default(&soap);
  if((soap_err=soap_call___nl2__add(&soap,client->SOAP_URL(url->Path().c_str()).c_str(),"add",&req,&r)) !=
SOAP_OK) {
    odlog(1)<<"Failed to pass information to database"<<std::endl;
    if(LogTime::level > 0) soap_print_fault(&soap, stderr);
    client->disconnect();
    return false;;
  } else if(r.result == NULL) {
    odlog(1)<<"Record refused."<<std::endl;
  } else if(r.result->Code != nl2__ResultCode__NoError) {
    odlog(1)<<"Record refused. Error code: "<<r.result->Code<<std::endl;
    return false;
  };
  return (r.result->Code == nl2__ResultCode__NoError);
}

bool LoggerClient::ReportV1(const char* url_,std::list<nl2__UsageRecord>& info) {
  if(!NewURL(url_)) return false;
  int r = 0;
  for(std::list<nl2__UsageRecord>::iterator i = info.begin();i!=info.end();++i) {
    // connect to server if not yet connected
    if(client->connect() != 0) {
      odlog(1)<<"Failed to connect to "<<*url<<std::endl;
      return false;
    };
    // Convert information to version 1
    nl__jobinfo i_;
    i_.start=i->submissiontime?*(i->submissiontime):0;
    i_.end=i->endtime?*(i->endtime):0;
    i_.cluster=(char*)(i->cluster?i->cluster->c_str():NULL);
    i_.user=(char*)(i->usersn.c_str());
    i_.id=(char*)(i->ngjobid.c_str());
    i_.name=(char*)(i->jobname?i->jobname->c_str():NULL);
    i_.failure=(char*)(i->failurestring?i->failurestring->c_str():NULL);
    i_.lrms=(char*)(i->lrms?i->lrms->c_str():NULL);
    i_.queue=(char*)(i->queue?i->queue->c_str():NULL);
    i_.rsl=(char*)(i->description?i->description->c_str():NULL);
    i_.ui=(char*)(i->clienthost?i->clienthost->c_str():NULL);
    i_.usedcpu=i->usedcputime?*(i->usedcputime):0;
    i_.usedmem=i->usedmemory?*(i->usedmemory):0;
    // Send report
    int soap_err = SOAP_OK;
    if((soap_err=soap_call_nl__add(&soap,client->SOAP_URL(url->Path().c_str()).c_str(),"add",&i_,r)) != SOAP_OK) {
      odlog(1)<<"Failed to pass information to database"<<std::endl;
      r=1;
      if(LogTime::level > 0) soap_print_fault(&soap, stderr);
      client->disconnect();
      break;
    } else if(r != 0) {
      odlog(1)<<"Record refused. Error code: "<<r<<std::endl;
      break;
    };
  };
  // Do not disconnect
  return (r==0);
}

bool LoggerClient::Query(const char* url_,const char* q,unsigned long long offset,std::list<nl2__UsageRecord>& info_) {
  // Try V2, if it fails try V1
  if(QueryV2(url_,q,offset,100,info_)) return true;
  return QueryV1(url_,q,offset,info_);
}

bool LoggerClient::QueryV1(const char* url_,const char* q,unsigned long long offset,std::list<nl2__UsageRecord>& info_) {
  array_jobinfo info;
  if(!NewURL(url_)) return false;
  // connect to server if not yet connected
  if(client->connect() != 0) {
    odlog(1)<<"Failed to connect to "<<*url<<std::endl;
    return false;
  };
  // Query server
  int soap_err = SOAP_OK;
  array_jobinfo r;
  if((soap_err=soap_call_nl__get(&soap,client->SOAP_URL(url->Path().c_str()).c_str(),"get",(char*)q,offset,r)) != SOAP_OK) {
    odlog(1)<<"Failed to get information from database"<<std::endl;
    if(LogTime::level > 0) soap_print_fault(&soap, stderr);
    client->disconnect();
    return false;
  };
  // Do not disconnect
  for(int i = 0; i<info.__size_job; i++) {
    nl2__UsageRecord job;
    copy_jobinfo(job,r.job[i],&soap);
    info_.push_back(job);
  };
  return true;
}

bool LoggerClient::QueryV2(const char* url_,const char* q,unsigned long long offset,unsigned long long size,std::list<nl2__UsageRecord>& info_) {
  if(!NewURL(url_)) return false;
  nl2__getRequest req;
  req.soap_default(&soap);
  req.query=soap_new_std__string(&soap,q);
  req.offset=offset;
  req.size=size;
  nl2__getResponse resp;
  resp.soap_default(&soap);
  int soap_err = SOAP_OK;
  if((soap_err=soap_call___nl2__get(&soap,client->SOAP_URL(url->Path().c_str()).c_str(),"get",&req,&resp))
                                                           != SOAP_OK) {
    odlog(1)<<"Failed to get information from database"<<std::endl;
    if(LogTime::level > 0) soap_print_fault(&soap, stderr);
    client->disconnect();
    return false;
  };
  for(std::vector<nl2__UsageRecord*>::iterator i = resp.job.begin();
                                        i!=resp.job.end();++i) {
    if(!*i) continue;
    info_.push_back(**i);
  };
  return true;
}

#define MAX_CLIENTS 10
HTTP_ClientSOAP* clients[MAX_CLIENTS];

typedef struct {
  bool start,end,cluster,user,id,name,failure,lrms,queue,rsl,ui,usedcpu,usedmem;
} elements_t;

#define set_element(v,e,name) if(strcmp(name,v) == 0) { e=true; }

void MaskElements(nl2__UsageRecord& j,const elements_t& e) {
  if(!e.start)   j.submissiontime=NULL;
  if(!e.end)     j.endtime=NULL;
  if(!e.usedcpu) j.usedcputime=NULL;
  if(!e.usedmem) j.usedmemory=0;
  if(!e.cluster) j.cluster=NULL;
  //if(!e.user) 
  //if(!e.id) 
  if(!e.name)    j.jobname=NULL;
  if(!e.failure) j.failurestring=NULL;
  if(!e.lrms)    j.lrms=NULL;
  if(!e.queue)   j.queue=NULL;
  if(!e.rsl)     j.description=NULL;
  if(!e.ui)      j.clienthost=NULL;
}

int logger(const char* url,char const * const * dirs,int num,time_t ex_period = 0);
int query(const char* url,char const * q,unsigned long int offset,unsigned int size,bool xml_out,const elements_t& e);
 
int main(int argc,char* argv[]) {
  char* url = NULL;
  elements_t e = { true, true, true, true, true, true, true, true, true, true, true, true, true };
  time_t ex_period = 0;

  opterr=0;
  bool q = false;
  bool xml_out = false;
  unsigned long int offset = 0;
  unsigned long int size = 100;
  int n;
  const char* basename = strrchr(argv[0],'/');
  if(basename == NULL) basename=argv[0];
  if(strcmp(basename,"nglogq") == 0) q=true;
  while((n=getopt(argc,argv,":hxe:qu:d:o:s:E:")) != -1) {
    switch(n) {
      case ':': { olog<<"Missing argument\n"; return 1; };
      case '?': { olog<<"Unrecognized option\n"; return 1; };
      case 'h': {
        std::cout<<"logger [-h] [-d level] [-u url] {control_dir ...|-q [-x] [-o offset] [-s size] [-e element,element,...] [-E expiration_period_days] [query]}"<<std::endl;
         return 0;
      };
      case 'u': {
        url=optarg;
      }; break;
      case 'q': {
        q=true;
      }; break;
      case 'd': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper debug level '"<<optarg<<"'"<<std::endl;
          exit(1);
        };
        LogTime::level=i;
      }; break;
      case 'e': {
        e.start=false; e.end=false; e.cluster=false; e.user=false;
        e.id=false; e.name=false; e.failure=false; e.lrms=false;
        e.queue=false; e.rsl=false; e.ui=false; e.usedcpu=false;
        e.usedmem=false;
        for(const char* p=optarg;*p;) {
          const char* pp = strchr(p,',');
          std::string name;
          if(pp) { name.assign(p,pp-p); } else { name.assign(p); };
          set_element(name.c_str(),e.start,"start")
          else set_element(name.c_str(),e.end,"end")
          else set_element(name.c_str(),e.cluster,"cluster")
          else set_element(name.c_str(),e.user,"user")
          else set_element(name.c_str(),e.id,"id")
          else set_element(name.c_str(),e.name,"name")
          else set_element(name.c_str(),e.failure,"failure")
          else set_element(name.c_str(),e.lrms,"lrms")
          else set_element(name.c_str(),e.queue,"queue")
          else set_element(name.c_str(),e.ui,"ui")
          else set_element(name.c_str(),e.usedcpu,"usedcpu")
          else set_element(name.c_str(),e.usedmem,"usedmem");
          if(!pp) break;
          p=pp+1;
        };
      }; break;
      case 'o': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper offset '"<<optarg<<"'"<<std::endl;
          exit(1);
        };
        offset=i;
      }; break;
      case 's': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<=0)) {
          olog<<"Improper size '"<<optarg<<"'"<<std::endl;
          exit(1);
        };
        size=i;
      }; break;
      case 'E': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<=0)) {
          olog<<"Improper expiration period '"<<optarg<<"'"<<std::endl;
          exit(1);
        };
        ex_period=i; ex_period*=(60*60*24);
      }; break;
      case 'x': {
        xml_out=true;
      }; break;
      default: { olog<<"Options processing error\n"; return 1; };
    };
  };
  if(xml_out) { e.start=true; e.end=true; };
  if(!q) return logger(url,argv+optind,argc-optind,ex_period);
  return query(url,argv[optind],offset,size,xml_out,e);
}

int logger(const char* url,char const * const * dirs,int num,time_t ex_period) {
  int ret = 0;
  if(globus_module_activate(GLOBUS_IO_MODULE) != GLOBUS_SUCCESS) exit(1);

  {
  LoggerClient client;
  for(int n = 0;dirs[n] && n<num;n++) {
    std::string logdir(dirs[n]); logdir+="/logs/";
    std::string logger_url(url?url:"");
    struct dirent file_;
    struct dirent *file;
    odlog(1)<<"Processing directory: "<<logdir<<std::endl;
    DIR *dir=opendir(logdir.c_str());
    if(dir == NULL) continue;
    for(;;) {
      readdir_r(dir,&file_,&file);
      if(file == NULL) break;
      std::string logfile = logdir+file->d_name;
      struct stat st;
      if(stat(logfile.c_str(),&st) != 0) continue;
      if(!S_ISREG(st.st_mode)) continue;
      odlog(1)<<"Processing file: "<<logfile<<std::endl;
      bool result = true;
      /* read file */
      char buf[4096];
      std::ifstream f(logfile.c_str());
      if(!f.is_open() ) continue; /* can't open file */
      JobRecord j(f);
      f.close();
      if(!j) {
        result=false;
        odlog(0)<<"Removing unreadable job information in "<<logfile<<std::endl;
        unlink(logfile.c_str());
      } else {
        if(j.url.length()) logger_url=j.url;
        if(logger_url.length() == 0) {
          odlog(2)<<"No service URL provided"<<std::endl;
          result = false;
        } else {
          std::list<nl2__UsageRecord> recs;
          recs.push_back(j.info);
          odlog(2)<<"Reporting to: "<<logger_url<<std::endl;
          result = client.Report(logger_url.c_str(),recs);
        };
        if(result) {
          odlog(1)<<"Passed information about job "<<j.info.ngjobid<<std::endl;
          unlink(logfile.c_str());
        } else {
          odlog(0)<<"Failed to pass information about job "<<
                  j.info.ngjobid<<std::endl;
          ret=-1;
          // Check creation time and remove it if really too old
          if((ex_period) && 
             (((unsigned int)(time(NULL)-st.st_mtime)) > ex_period)) {
            odlog(0)<<"Removing outdated information about job "<<
                    j.info.ngjobid<<std::endl;
            unlink(logfile.c_str());
          };
        };
      };
    };
    closedir(dir);
  };
  };
  globus_module_deactivate(GLOBUS_IO_MODULE);
  return ret;
}

int query(const char* url,char const * q,unsigned long int offset,unsigned int size,bool xml_out,const elements_t& e) {
  if(url == NULL) {
    odlog(0)<<"URL is needed"<<std::endl;
    return 1;
  };
  if(globus_module_activate(GLOBUS_IO_MODULE) != GLOBUS_SUCCESS) exit(1);
  std::list<nl2__UsageRecord> info;
  bool res = true;
  unsigned long int o = offset;
  unsigned int s = 0;
  {
    LoggerClient client;
    for(;s<size;) {
      res = client.Query(url,q,o,info);
      if(!res) break;
      if(info.size() == s) break;
      s+=info.size(); o+=info.size();
    };
    if(!res && !s) {
      odlog(0)<<"Failed to retrieve information"<<std::endl;
    } else if(s) {
      if(!res) odlog(0)<<"Warning: not full information was retrieved"<<std::endl;
      res=true;
      int i = 0;
      for(std::list<nl2__UsageRecord>::iterator i=info.begin();
                                                 i!=info.end();i++) {
        MaskElements(*i,e);
      };
      if(!xml_out) {
        for(std::list<nl2__UsageRecord>::iterator i=info.begin();
                                                   i!=info.end();i++) {
          std::cout<<*i;
        };
      } else { // serialize using gSOAP
        struct soap sp;
        soap_init(&sp);
        sp.namespaces=logger2_soap_namespaces;
        for(std::list<nl2__UsageRecord>::iterator i=info.begin();
                                                   i!=info.end();i++) {
          soap_begin(&sp);
          soap_set_omode(&sp,SOAP_XML_GRAPH | SOAP_XML_CANONICAL);
          i->soap_serialize(&sp);
          sp.os=&std::cout;
          soap_begin_send(&sp);
          i->soap_put(&sp,"job","nl2::UsageRecord");
          soap_end_send(&sp);
          soap_end(&sp);
        };
        soap_done(&sp);
      };
    } else {
      odlog(0)<<"Warning: no data was retrieved"<<std::endl;
    };
  };
  globus_module_deactivate(GLOBUS_IO_MODULE);
  if(!res) return 1;
  return 0;
}
